(* Documentation for the col package: a syntax extension of OCaml
   for the safe manipulation of flat records and their conversion
   from and to text files (CSV), tuples, objects and raw string arrays.

   The advantages over using a syntax extension are the following:
   - records can be defined with default values
   - a "create" function is automatically defined; its arguments are labeled,
     and the ones with default values are optional.
   - the equivalent functions are provided for objects and tuples and one can 
     convert one type into another.
   - record fields of the same type can be selected at runtime.
     For instance, the following "get" function takes a record and 
     returns a float although the record contains fields of types 
     other than float:
        let get = type_float_field (if cond then "x" else "y")
   - a list of records/objects/tuples can be read safely from CSV files with
     a header, even if the columns are in any order and if unknown columns
     are present.
   - if the CSV file has no header, the columns are assumed to be in the order
     of the type definition. Additional columns are ignored.
   - plain English names can be given to columns labels, so that the CSV
     files are ready for being manipulated in your favorite spreadsheet 
     or plotting program.
   - conversion of polymorphic variants without argument to and from strings
     is fully automatic

   Below (file <a href="#doc.ml">doc.ml</a>) is a sample of code using the syntax extension.
   <a href="#doc.mli">doc.mli</a> shows all the type definitions, submodules
   and functions that are created. 
   <a href="#doc.ppo">doc.ppo</a> is the same code, after expansion by camlp4
   into regular OCaml.


   Compile with:
   ocamlfind ocamlopt -c yourprogram.ml -syntax camlp4o -package col
*)


(* We define a custom type that we will use in record fields
   (no syntax novelty here) *)
module Date = 
struct
  type t = int * int
  let compare (m1, y1) (m2, y2) = 
    let c = compare y1 y2 in
    if c <> 0 then c else compare m1 m2
  let of_string s = Scanf.sscanf s "%i-%i" (fun m y -> (m, y))
  let to_string (m, y) = Printf.sprintf "%02i-%i" m y
end

(* A special syntax for defining a module which has the same structure
   as the Date module above. See <a href="#doc.mli">doc.mli</a>.
   It defines a "compare" function which uses the order in which
   the tags appear in the type declaration. *)
type tag fruit = [ `Apple | `Orange | `Banana | `Passion "Passion fruit" ]

(* The main syntax extension: it defines a record type of the same name,
   plus the equivalent tuple and object types, and functions to manipulate
   them. See <a href="#doc.mli">doc.mli</a> for the interface
   and <a href="#doc.ppo">doc.ppo</a> for the expanded OCaml code. *)
type col example = {
  alpha; (* field type defaults to string (builtin type) *)
  bravo "Bravo!"; (* use another label in data files *)
  charlie = "empty"; (* default value *)
  delta : string; (* field type specification *)
  echo "echo echo" : string = "nothing"; (* combining the previous features *)
  foxtrot : int; (* another builtin type *)
  golf : int option;
  hotel : int option = Some 123;
  india : float; (* another builtin type *)
  juliet : float option;
  kilo : bool; (* another builtin type *)
  lima : bool option;
  mike "date (mm-yyyy)" : Date; (* a custom type given as a module *)
  november : Date option = None;
  oscar : [ `Apple | `Orange ]; (* inline tag definition *)
  papa : [ `Apple | `Banana ] = `Banana;
  quebec : Fruit = `Passion; (* external tag definition *)
  mutable romeo : int; (* mutable for records and objects, not tuples *)
  (* sierra
     tango
     uniform
     victor
     whisky
     x-ray
     yankee
     zulu *)
}

(* An illustration of how to reuse record and variant types 
   that are already defined.

   The Unix.stat type contains information about a given file. We want to
   display this information for all the files of a directory, and sort
   them the way we like.

   The following example is a program which takes a directory as argument,
   and displays the full stat information on each file of the directory.
   The rows are explicitely sorted so that directories (S_DIR) come before 
   symbolic links (S_LNK) and regular files (S_REG).
*)

(* We wrap this into a module to make things clear, but it's not necessary. *)
module Test_predef =
struct
  (* Opening the module which provides our reused types 
     stat and file_kind is necessary: *)
  open Unix

  (* The order of the tags is used to define an appropriate "compare" 
     function (File_kind.compare). 
     The "predefined" keyword must be used to avoid redefining the type. *)
  type tag file_kind = predefined
    | S_CHR
    | S_BLK
    | S_FIFO
    | S_SOCK
    | S_DIR
    | S_LNK
    | S_REG

  (* st_kind is the only field which is not trivial. We use the File_kind
     module which is created by the "type tag" definition above. *)
  type col stats = predefined {
    st_dev : int;
    st_ino : int;
    st_kind : File_kind (*Unix.file_kind*);
    st_perm : int (*Unix.file_perm*);
    st_nlink : int;
    st_uid : int;
    st_gid : int;
    st_rdev : int;
    st_size : int;
    st_atime : float;
    st_mtime : float;
    st_ctime : float;
  }

  (* Now we use the functions from the Stats module 
     which has just been created. *)

  (* output functions *)
  let print_list l =
    print_endline (String.concat " " l)
  let print_row (file, stats) =
    print_list (file :: (Array.to_list (Stats.to_array stats)))

  (* reading the files and their stat records *)
  let dir = match Sys.argv with
      [| _; dir |] -> dir
    | _ -> Sys.getcwd ()
  let files = Array.to_list (Sys.readdir dir)
  let stats = 
    List.map (fun file -> (file, stat (Filename.concat dir file))) files

  (* Comparison function for stat records: 
     the fields to use are given from the most important
     to the least important, with the direction. `Up means the
     regular (ascending) direction, while `Down means the opposite 
     (descending). *)
  let cmp = Stats.compare_fields [`Up "st_kind"; `Down "st_size"]

  (* As a last resort, we use the file names to sort them in case of a tie, 
     and unfortunately they are not contained in the stat record
     so we have to write that code: *)
  let sort l =
    List.sort (fun (file1, stat1) (file2, stat2) ->
		 let c = cmp stat1 stat2 in
		 if c <> 0 then c else String.compare file1 file2)
      l
  
  (* Finally sort and print the result *)  
  let _ =
    (* header *)
    print_list ("file" :: (Array.to_list Stats.fields));
    (* data *)
    List.iter print_row (sort stats)
end 
